"""Set of checker functions for Zuul Lint."""

from __future__ import annotations

import pathlib


def check_job_playbook_paths(
    job: dict[str, dict | list[str | dict]]
) -> list[str]:
    """Check that all playbooks in a job have a valid path.

    Args:
    ----
        job: A dictionary containing a Zuul job.

    Returns:
    -------
        A list of invalid playbook paths.
    """
    invalid_paths = []

    for key in ["pre-run", "run", "post-run", "cleanup-run"]:
        paths = job.get(key)
        if paths is None:
            continue

        # Convert to list if it's a single object
        if not isinstance(paths, list):
            paths = [paths]

        for path in paths:
            if isinstance(path, str):
                if not pathlib.Path(path).exists():
                    invalid_paths.append(path)
            if isinstance(path, dict):
                if "name" in path:
                    if not pathlib.Path(path["name"]).exists():
                        invalid_paths.append(path["name"])

    return invalid_paths


def check_duplicated_jobs(
    jobs: dict[pathlib.Path, list[dict | None]]
) -> set[dict[str, str] | None]:
    """Check that all jobs are unique in different Zuul YAML files.

    Args:
    ----
        jobs: A dictionary containing a list of Zuul jobs.

    Returns:
    -------
        A set of duplicated jobs.
    """
    seen_items = set()
    duplicated_items = set()
    unique_items = set()

    for joblist in jobs.values():
        try:
            sublist_set = set(job["job"]["name"] for job in joblist)
        except KeyError:
            continue
        for job in sublist_set:
            if job in seen_items:
                duplicated_items.add(job)
            seen_items.add(job)
        unique_items.update(sublist_set)

    return duplicated_items


def check_inexistent_nodesets(
    nodesets: list[dict],
    jobs: list[dict | None],
) -> list[dict[str, str]]:
    """Check that all used nodesets exist.

    Args:
    ----
        nodesets: A list of nodesets.
        jobs: A list of Zuul jobs.

    Returns:
    -------
        A list of inexistent nodesets.
    """
    nodeset_list = set()
    for nodeset in nodesets:
        nodeset_list.add(nodeset["nodeset"]["name"])
        try:
            node_list = nodeset["nodeset"]["nodes"]
        except KeyError:
            continue
        for node in node_list:
            if isinstance(node["name"], str):
                nodeset_list.add(node["name"])
            if isinstance(node["name"], list):
                for node_name in node["name"]:
                    nodeset_list.add(node_name)
    inexistent_nodesets = set()

    for job in jobs:
        try:
            if isinstance(job["job"]["nodeset"], str):
                nodeset = {}
                nodeset["name"] = job["job"]["nodeset"]
                job_nodeset_list = [nodeset]
            else:
                job_nodeset_list = job["job"]["nodeset"]["nodes"]
        except KeyError:
            continue
        for nodeset in job_nodeset_list:
            try:
                if nodeset["name"] not in nodeset_list:
                    inexistent_nodesets.add(nodeset["name"])
            except TypeError:
                if job_nodeset_list.get("name", None) and job_nodeset_list.get("name", None) not in nodeset_list:
                    inexistent_nodesets.add(job_nodeset_list["name"])

    return inexistent_nodesets



"""ZuuL Lint.

A linter for Zuul configuration files.
"""

from __future__ import annotations

import argparse
import importlib.metadata
import pathlib
import sys
from collections import defaultdict

import yaml
from jsonschema import Draft201909Validator

import zuulcilint.checker as zuul_checker
import zuulcilint.utils as zuul_utils
from zuulcilint.utils import MsgSeverity, ZuulObject

# Register custom yaml constructor for "encrypted/pkcs1-oaep"
yaml.SafeLoader.add_constructor(
    "!encrypted/pkcs1-oaep",
    zuul_utils.encrypted_pkcs1_oaep_constructor,
)


def lint(file_path: str, schema: dict) -> int:
    """Validate a YAML file against a JSON schema.

    Args:
    ----
        file_path: A string representing the path to the YAML file to validate.
        schema: A JSON schema to validate against.

    Returns:
    -------
        The number of validation errors encountered.
    """
    print(f"{file_path}")
    errors = 0
    validator = Draft201909Validator(schema)

    try:
        with pathlib.Path.open(pathlib.Path(file_path), encoding="utf-8") as yaml_in:
            try:
                obj = yaml.safe_load(yaml_in)
                va_errors = validator.iter_errors(obj)
                for e in va_errors:
                    zuul_utils.print_bold("Validation error:", MsgSeverity.ERROR)
                    print(f"File: {file_path}")
                    print(f"Message: {e.message}")
                    print(f"Path: {list(e.path)}")
                    print(f"Schema Path: {list(e.schema_path)}\n")
                    errors += 1
            except yaml.YAMLError as e:
                print(f"YAML Parse Error: {e}")
                errors += 1
    except FileNotFoundError as e:
        print(f"{e.filename} not found!\nExiting")
        sys.exit(1)

    return errors


def lint_single_yaml_file(file_path: pathlib.Path, schema: dict) -> int:
    """Lint a single Zuul YAML file.

    Args:
    ----
        file_path: A string representing the path to the YAML file to validate.
        schema: A JSON schema to validate against.

    Returns:
    -------
        The number of validation errors encountered.
    """
    return lint(file_path, schema=schema)


def lint_all_yaml_files(file_paths: list[pathlib.Path], schema: dict) -> int:
    """Lint all Zuul YAML files.

    Args:
    ----
        file_paths: A list of strings representing the paths to the YAML files to validate.
        schema: A JSON schema to validate against.

    Returns:
    -------
        The number of validation errors encountered.
    """
    return sum(lint_single_yaml_file(file_path, schema) for file_path in file_paths)


def lint_playbook_paths(zuul_yaml_files: list[pathlib.Path]) -> list[str]:
    """Lint playbook paths in all Zuul YAML files.

    Args:
    ----
        zuul_yaml_files: A list of Zuul YAML files.

    Returns:
    -------
        A list of invalid playbook paths.
    """
    invalid_paths = []
    for file_path in zuul_yaml_files:
        jobs = zuul_utils.get_zuul_object_from_yaml(ZuulObject.JOB, file_path)
        for job in jobs:
            invalid_paths.extend(
                zuul_checker.check_job_playbook_paths(job.get(ZuulObject.JOB.value, {})),
            )
    return invalid_paths


def get_all_zuul_yaml_files(files: list[str]) -> list[pathlib.Path]:
    """Get all Zuul YAML/YML files from the specified file(s) or path(s).

    Args:
    ----
        files: A list of strings representing the file(s) or path(s) to lint.

    Returns:
    -------
        A list of Zuul YAML/YML files.
    """
    zuul_yaml_files = defaultdict(list)
    for file in files:
        for file_type, paths in zuul_utils.get_zuul_yaml_files(pathlib.Path(file)).items():
            zuul_yaml_files[file_type].extend(paths)

    return zuul_yaml_files


def get_all_zuul_objects_by_type(
    zuul_yaml_files: list[pathlib.Path],
    zuul_obj: ZuulObject,
) -> list[dict]:
    """Get all Zuul objects from provided Zuul YAML files.

    Args:
    ----
        zuul_yaml_files: A list of Zuul YAML files.
        zuul_obj: A ZuulObject enum.

    Returns:
    -------
        A list of Zuul objects.
    """
    all_zuul_objects = []
    for file_path in zuul_yaml_files:
        zuul_objects = zuul_utils.get_zuul_object_from_yaml(zuul_obj, file_path)
        all_zuul_objects.extend(zuul_objects)
    return all_zuul_objects


def print_warnings(
    warnings: dict,
    severity: MsgSeverity = MsgSeverity.WARNING,
) -> None:
    """Print warnings.

    Args:
    ----
        warnings: A dictionary containing warnings.
        severity: A MsgSeverity enum.

    Returns:
    -------
        None.
    """
    n_bad_yaml = len(warnings["warnings"]["bad_yaml_files"])
    n_duplicate = len(warnings["warnings"]["duplicated_jobs"])
    n_nodeset = len(warnings["warnings"]["inexistent_nodesets"])

    if n_bad_yaml == 0 and n_duplicate == 0 and n_nodeset == 0:
        return

    if severity == MsgSeverity.WARNING:
        zuul_utils.print_bold("Warnings", MsgSeverity.WARNING)
        print(f"Total {severity.value}s: {n_duplicate + n_duplicate + n_nodeset}")

    if n_bad_yaml:
        zuul_utils.print_bold(f"File extension {severity.value}s:", severity)
        zuul_utils.print_bold(
            f"Found {n_bad_yaml} files with 'yml' extension",
            None,
        )
        for file_path in warnings["warnings"]["bad_yaml_files"]:
            print(f"{file_path}")

    if n_duplicate:
        zuul_utils.print_bold(f"Duplicate job {severity.value}s:", severity)
        zuul_utils.print_bold(f"Found {n_duplicate} duplicate jobs", None)
        for job in warnings["warnings"]["duplicated_jobs"]:
            print(f"{job}")

    if n_nodeset:
        zuul_utils.print_bold(f"Inexistent nodeset {severity.value}s:", severity)
        zuul_utils.print_bold(f"Found {n_nodeset} inexistent nodesets", None)
        for nodeset in warnings["warnings"]["inexistent_nodesets"]:
            print(f"{nodeset}")


def print_results(
    results: dict,
    warnings_as_errors,
    ignore_warnings,
) -> None:
    """Print the linting results.

    Args:
    ----
        results: A dictionary containing linting results.
        warnings_as_errors: A boolean indicating whether to handle warnings as errors.
        ignore_warnings: A boolean indicating whether to ignore warnings.

    Returns:
    -------
        None.
    """
    duplicated_jobs = results["warnings"]["duplicated_jobs"]
    inexistent_nodesets = results["warnings"]["inexistent_nodesets"]
    bad_yaml_files = results["warnings"]["bad_yaml_files"]
    total_yaml_errors = results["errors"]["yaml"]
    total_playbook_path_errors = results["errors"]["playbook_paths"]
    total_warnings = len(bad_yaml_files) + len(duplicated_jobs) + len(inexistent_nodesets)
    total_errs = total_yaml_errors + total_playbook_path_errors
    extra_msg = ""

    # --warnings-as-errors flag has higher precedence than --ignore-warnings
    if warnings_as_errors:
        total_errs += total_warnings
        if bad_yaml_files:
            extra_msg += f"\nFile extension errors: {len(bad_yaml_files)}"
        if duplicated_jobs:
            extra_msg += f"\nDuplicated jobs errors: {len(duplicated_jobs)}"
        if inexistent_nodesets:
            extra_msg += f"\nInexistent nodesets errors: {len(inexistent_nodesets)}"
        print_warnings(warnings=results, severity=MsgSeverity.ERROR)
    elif not ignore_warnings:
        print_warnings(warnings=results)

    if total_errs == 0:
        zuul_utils.print_bold("Passed", MsgSeverity.SUCCESS)
        sys.exit(0)

    zuul_utils.print_bold("Failed", MsgSeverity.ERROR)
    err_msg = f"Total errors: {total_errs}\n"

    if total_yaml_errors:
        err_msg += f"YAML validation errors: {total_yaml_errors}"
    if total_playbook_path_errors:
        err_msg += f"\nPlaybook path errors: {total_playbook_path_errors}"

    zuul_utils.print_bold(f"{err_msg + extra_msg}", MsgSeverity.ERROR)
    sys.exit(1)


def main():
    """Parse command-line arguments and run the Zuul linter on the specified file(s).

    Returns
    -------
        None.
    """
    parser = argparse.ArgumentParser(prog="zuulcilint")
    parser.add_argument(
        "--version",
        action="version",
        version=importlib.metadata.version("zuulcilint"),
    )
    parser.add_argument(
        "--check-playbook-paths",
        "-c",
        help="check that playbook paths are valid",
        action="store_true",
    )
    parser.add_argument("file", nargs="+", help="file(s) or paths to lint")
    parser.add_argument(
        "--schema",
        "-s",
        help="path to Zuul schema file",
        default=pathlib.Path(__file__).parent / "zuul-schema.json",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--ignore-warnings",
        "-i",
        help="ignore warnings",
        action="store_true",
    )
    parser.add_argument(
        "--warnings-as-errors",
        help="handle warnings as errors",
        action="store_true",
    )

    args = parser.parse_args()
    schema = zuul_utils.get_zuul_schema(schema_file=args.schema)
    all_zuul_yaml_files = get_all_zuul_yaml_files(args.file)
    zuul_good_yaml = all_zuul_yaml_files["good_yaml"]
    zuul_bad_yaml = all_zuul_yaml_files["bad_yaml"]

    # Initialize results dictionary
    results = {
        "errors": {"yaml": 0, "playbook_paths": 0},
        "warnings": {
            "duplicated_jobs": [],
            "inexistent_nodesets": [],
            "bad_yaml_files": zuul_bad_yaml,
        },
    }

    # Lint all Zuul YAML files
    results["errors"]["yaml"] = lint_all_yaml_files(zuul_good_yaml, schema)

    # Check playbook paths if specified
    if args.check_playbook_paths:
        zuul_utils.print_bold("Checking playbook paths", MsgSeverity.INFO)
        invalid_playbook_paths = lint_playbook_paths(zuul_good_yaml)
        if invalid_playbook_paths:
            results["errors"]["playbook_paths"] = len(invalid_playbook_paths)
            zuul_utils.print_bold("Invalid playbook paths:", MsgSeverity.ERROR)
            for path in invalid_playbook_paths:
                print(f"{path}")
        else:
            print("No invalid playbook paths")

    # Check duplicated jobs
    zuul_utils.print_bold("Checking for duplicate jobs", MsgSeverity.INFO)
    jobs_dict = {}
    for yaml_file in zuul_good_yaml:
        jobs_dict[yaml_file] = get_all_zuul_objects_by_type([yaml_file], ZuulObject.JOB)

    duplicated_jobs = zuul_checker.check_duplicated_jobs(jobs_dict)

    if duplicated_jobs:
        for job in duplicated_jobs:
            print(f"{job}")
    else:
        print("No duplicate jobs found")
    results["warnings"]["duplicated_jobs"] = duplicated_jobs

    # Check for inexistent nodesets
    zuul_utils.print_bold("Checking for inexistent nodesets", MsgSeverity.INFO)
    inexistent_nodesets = zuul_checker.check_inexistent_nodesets(
        get_all_zuul_objects_by_type(zuul_good_yaml, ZuulObject.NODESET),
        get_all_zuul_objects_by_type(zuul_good_yaml, ZuulObject.JOB),
    )
    if inexistent_nodesets:
        for nodeset in inexistent_nodesets:
            print(f"{nodeset}")
    else:
        print("No inexistent nodesets found")
    results["warnings"]["inexistent_nodesets"] = inexistent_nodesets

    # Print results
    print_results(
        results,
        args.warnings_as_errors,
        args.ignore_warnings,
    )


if __name__ == "__main__":
    main()



"""Set of utility functions for Zuul Lint."""

from __future__ import annotations

import json
import pathlib
import sys
from collections import defaultdict
from enum import Enum

import yaml


class MsgSeverity(Enum):

    """Enum for message severity."""

    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    SUCCESS = "success"


class MsgTypeColor(Enum):

    """Enum for message type colors."""

    ERROR = "\33[1;49;31m"
    WARNING = "\33[1;49;33m"
    INFO = "\33[1;49;34m"
    SUCCESS = "\33[1;49;32m"
    DEFAULT = "\33[1;49;37m"
    RESET = "\33[0m"


class ZuulObject(Enum):

    """Enum representing Zuul objects."""

    JOB = "job"
    NODESET = "nodeset"
    PIPELINE = "pipeline"
    PRAGMA = "pragma"
    PROJECT = "project"
    QUEUE = "queue"
    SECRET = "secret" # noqa: S105
    SEMAPHORE = "semaphore"
    TEMPLATE = "project-template"


def get_zuul_schema(schema_file: str) -> dict:
    """Load the Zuul schema from a JSON file.

    Args:
    ----
        schema_file: The path to the JSON schema file.

    Returns:
    -------
        A dictionary representing the Zuul schema.
    """
    try:
        with pathlib.Path.open(schema_file, encoding="utf-8") as f:
            return json.load(f)
    except FileNotFoundError:
        print(f"Schema file not found: {schema_file}", file=sys.stderr)
        sys.exit(1)
    except json.JSONDecodeError:
        print(f"Invalid JSON in schema file: {schema_file}", file=sys.stderr)
        sys.exit(1)


def get_zuul_yaml_files(path: pathlib.Path) -> dict[str, list[pathlib.Path]]:
    """Retrieve a dictionary of Zuul YAML/YML files from the specified path.

    Args:
    ----
        path (Path): The path to search for Zuul YAML files.

    Returns:
    -------
        dict[str, List[Path]]: A dictionary containing the keys 'good_yaml'(.yaml) and
        'bad_yaml'(.yml), where values are lists containing the Path objects for each
        of the file extensions found.
    """
    zuul_yaml_files = defaultdict(list)

    if path.is_file():
        if path.suffix == ".yaml":
            zuul_yaml_files["good_yaml"].append(path)
        elif path.suffix == ".yml":
            zuul_yaml_files["bad_yaml"].append(path)
    elif path.is_dir():
        for p in path.iterdir():
            for file_type, yaml_file_path in get_zuul_yaml_files(p).items():
                zuul_yaml_files[file_type].extend(yaml_file_path)

    else:
        print(f"Skipping {path}")

    return zuul_yaml_files


def get_zuul_object_from_yaml(
    obj_type: ZuulObject, zuul_yaml_file: str,
) -> list[dict[str, str] | None]:
    """Retrieve a list of Zuul objects from the specified YAML file.

    Args:
    ----
        obj_type: The type of object to search for.
        zuul_yaml_file: The path to the YAML file to search for Zuul objects.

    Returns:
    -------
        A list of dictionaries representing the Zuul objects found.
    """
    try:
        with pathlib.Path.open(zuul_yaml_file, encoding="utf-8") as f:
            return [obj for obj in yaml.safe_load(f) if obj.get(obj_type.value)]
    except FileNotFoundError:
        print(f"Zuul YAML file not found: {zuul_yaml_file}", file=sys.stderr)
        sys.exit(1)
    except yaml.YAMLError:
        print(f"Invalid YAML in Zuul YAML file: {zuul_yaml_file}", file=sys.stderr)
        sys.exit(1)
    except AttributeError:
        print(f"Invalid Zuul YAML file detected: {obj_type}", file=sys.stderr)
        sys.exit(1)


def get_playbook_paths_from_job(job: dict[str, str] | None) -> list[str | None]:
    """Retrieve a list of playbook paths from the specified job.

    Args:
    ----
        job: A dictionary representing a Zuul job.

    Returns:
    -------
        A list of playbook paths.
    """
    path_keys = ["pre-run", "run", "post-run"]

    return [job.get(key) for key in path_keys if job.get(key)]


def get_files_with_extension(path: str, extension: str) -> list[pathlib.Path]:
    """Get files with provided extension from provided path and it's subdirectories.

    Args:
    ----
        path: path to directory to search
        extension: file extension to search for

    Returns:
    -------
        A list of files with provided extension.
    """
    return list(pathlib.Path(path).rglob(f"*.{extension}"))


def encrypted_pkcs1_oaep_constructor(loader: str, node: str) -> str:
    """Handle encrypted strings in YAML files.

    Args:
    ----
        loader: The YAML loader.
        node: The YAML node.

    Returns:
    -------
        The decrypted string.
    """
    if isinstance(node, yaml.ScalarNode):
        # Handle scalar node
        return loader.construct_scalar(node)
    if isinstance(node, yaml.SequenceNode):
        # Handle sequence node
        return loader.construct_sequence(node)

    # Handle other node types if needed
    raise yaml.constructor.ConstructorError(
        None,
        None,
        "invalid node type: '%s'" % type(node),
        node.start_mark,
    )


def print_bold(msg: str, msg_type: MsgSeverity) -> None:
    """Print a bold message.

    Args:
    ----
        msg: The message to print.
        msg_type: The type of message to print.

    Returns:
    -------
        None.
    """
    if msg_type == MsgSeverity.ERROR:
        print(f"\n{MsgTypeColor.ERROR.value}{msg}{MsgTypeColor.RESET.value}", file=sys.stderr)
    elif msg_type == MsgSeverity.WARNING:
        print(f"\n{MsgTypeColor.WARNING.value}{msg}{MsgTypeColor.RESET.value}")
    elif msg_type == MsgSeverity.INFO:
        print(f"\n{MsgTypeColor.INFO.value}{msg}{MsgTypeColor.RESET.value}")
    elif msg_type == MsgSeverity.SUCCESS:
        print(f"\n{MsgTypeColor.SUCCESS.value}{msg}{MsgTypeColor.RESET.value}")
    else:
        print(f"{MsgTypeColor.DEFAULT.value}{msg}{MsgTypeColor.RESET.value}")
